/*
	Copyright 2002 Kevin O'Donnell

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * Include the core server components.
 */
#include "apr_strings.h"
#include "apr_thread_proc.h"
#include "apr_ring.h"
#include "apr_time.h"
#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_log.h"
#include "util_script.h"
#include <stdlib.h>
#include <stdio.h>


#define unless(c) if(!(c))

/**
 * The max number of child process arguments. Normally, 64 is enough.
 */
 #define APP_MAX_ARGC	64


/********************************
 * Initialize some stuff
 ********************************/

/* apache module name */
module AP_MODULE_DECLARE_DATA prov_module;

/* A single cache entry */
typedef struct Prov_cache_element {
	APR_RING_ENTRY(Prov_cache_element) link;
    apr_time_t	entry_time;
    char *	cookie_value;
} Prov_cache_element;

// Define the struct for the head of the whole cache
 APR_RING_HEAD(Prov_cache_list, Prov_cache_element);

/* configuration structure */
typedef struct strProv_config_rec {
    char *	szProv_script;
    char *	szProv_cookiename;
    int 	nProv_chkcookie;
//    struct 	Prov_cache_list *Prov_cache_list_head;
 } strProv_config_rec;

typedef struct strProv_module_config_rec {
	apr_pool_t *cache_pool;
	struct Prov_cache_list *cache_list_head;
} strProv_module_config_rec;

 //struct Prov_cache_list Prov_cache_list_head, *ringp;
 //struct Prov_cache_list Prov_cache_list_head;

 char * cmd_argument_string;

 /*********************************************************
  *
  * Execute a child process, and wait for its termination.
  *
  *********************************************************/
 static apr_status_t run_program(const char *progname, const apr_procattr_t *pattr, apr_pool_t *mp, const char* const *env)
 {
	 fprintf(stderr,"mod_prov: FUNCTION call run_program\n");
	 fflush(stderr);

     apr_status_t rv;
     int argc = 0;
     const char* argv[APP_MAX_ARGC];
     apr_proc_t proc;

     fprintf(stderr,"mod_prov: Starting run_program\n");
     fflush(stderr);


     argv[argc++] = progname;
     /* you can specify command line options as follows,
        argv[argc++] = "-i";
        argv[argc++] = "argument"; */
     argv[argc++] = NULL;/* @argvs should be null-terminated */

     if ((rv = apr_proc_create(&proc, progname, (const char* const*)argv,
                               env, (apr_procattr_t*)pattr, mp)) != APR_SUCCESS) {
         return rv;
     }

     {
         int st;
         apr_exit_why_e why;

         rv = apr_proc_wait(&proc, &st, &why, APR_WAIT);
         if (APR_STATUS_IS_CHILD_DONE(rv)) {
             //printf("child done: why = %d, exit status = %d\n", why, st);
             fprintf(stderr,"mod_prov: child done: why = %d, exit status =%d\n",why, st);
             fflush(stderr);
         } else {
             //printf("child notdone\n");
             fprintf(stderr,"mod_prov: child not done\n");
             fflush(stderr);
         }
     }
     return APR_SUCCESS;
 }


 /***********************************************
  *
  * Function to add element to cache
  *
  ***********************************************/

 static struct Prov_cache_element *elt_new(strProv_module_config_rec *mod_conf)
 {
	 // This function initializes a cache element. Not totally clear what
	 // everything means, though..

	 fprintf(stderr,"mod_prov: CACHE: FUNCTION call elt_new\n");
	 fflush(stderr);

	 struct Prov_cache_element *elt;

     if (APR_RING_EMPTY(mod_conf->cache_list_head, Prov_cache_element, link)) {
    	 fprintf(stderr,"mod_prov: CACHE: is empty\n");
    	 fflush(stderr);
         elt = apr_pcalloc(mod_conf->cache_pool, sizeof(*elt));
         if (NULL == elt) {
             return NULL;
         }
     }
     else {
    	 fprintf(stderr,"mod_prov: CACHE: not empty\n");
    	 fflush(stderr);
         elt = APR_RING_FIRST(mod_conf->cache_list_head);
         APR_RING_REMOVE(elt, link);
     }

     APR_RING_ELEM_INIT(elt, link);
     elt->cookie_value = NULL;
     elt->entry_time = 0;
     return elt;
 }



 /************************************************************************
   *
   * Iterate function to print all entries in a table
   *
   ************************************************************************/
 int iterate_func(void *req, const char *key, const char *value) {
	 fprintf(stderr,"mod_prov: FUNCTION call iterate_func\n");
	 fflush(stderr);

//     int stat;
     char *line;
     request_rec *r = (request_rec *)req;
     if (key == NULL || value == NULL || value[0] == '\0')
         return 1;

     // Fill command argument string with all "key=value" combinations
     cmd_argument_string = apr_pstrcat(r->pool, cmd_argument_string, " ", key, "=", value, NULL);

     line = apr_psprintf(r->pool, "%s => %s\n", key, value);

     fprintf(stderr,"mod_prov: command arguments are %s\n",cmd_argument_string);
     fflush(stderr);

     return 1;
 }

int print_iterate_func(void *req, const char *key, const char *value) {
    char *line;
    request_rec *r = (request_rec *)req;
    if (key == NULL || value == NULL || value[0] == '\0')
        return 1;
   
    line = apr_psprintf(r->pool, "%s => %s\n", key, value);

    ap_log_rerror(APLOG_MARK,APLOG_WARNING|APLOG_NOERRNO, 0,r, "modprov: %s", line);

    return 1;
}


 /************************************************************************
  *
  * Look through 'Cookie' header for indicated cookie; extract it
  * and URL-unescape it. Return the cookie on success, NULL on failure
  *
  ************************************************************************/
 static char * extract_cookie(request_rec *r, const char *szCookie_name)
 {
	 fprintf(stderr,"mod_prov: FUNCTION call extract_cookie\n");
	 fflush(stderr);

	 char *szRaw_cookie_start=NULL, *szRaw_cookie_end;
	 char *szCookie;
	 /* get cookie string */
	 char *szRaw_cookie = (char*)apr_table_get(r->headers_in, "Cookie");
	 fprintf(stderr,"mod_prov: Raw cookies: %s\n",szRaw_cookie);
	 fflush(stderr);
	 unless(szRaw_cookie) return 0;

	 /* loop to search cookie name in cookie header */
	 do {
		 /* search cookie name in cookie string */
		 unless (szRaw_cookie =strstr(szRaw_cookie, szCookie_name)) return 0;
		 szRaw_cookie_start=szRaw_cookie;
		 /* search '=' */
		 unless (szRaw_cookie = strchr(szRaw_cookie, '=')) return 0;
	 } while (strncmp(szCookie_name,szRaw_cookie_start,szRaw_cookie-szRaw_cookie_start)!=0);

	 /* skip '=' */
	 szRaw_cookie++;

	 /* search end of cookie name value: ';' or end of cookie strings */
	 unless ((szRaw_cookie_end = strchr(szRaw_cookie, ';')) || (szRaw_cookie_end = strchr(szRaw_cookie, '\0'))) return 0;

	 /* dup the value string found in apache pool and set the result pool ptr to szCookie ptr */
	 unless (szCookie = apr_pstrndup(r->pool, szRaw_cookie, szRaw_cookie_end-szRaw_cookie)) return 0;
	 /* unescape the value string */
	 unless (ap_unescape_url(szCookie) == 0) return 0;

	 return szCookie;
 }


/*
 * This function is registered as a handler for HTTP methods and will
 * therefore be invoked for all GET requests (and others).  Regardless
 * of the request type, this function simply sends a message to
 * STDERR (which httpd redirects to logs/error_log).  A real module
 * would do *alot* more at this point.
 */
static int mod_prov_method_handler (request_rec *r)
{
	fprintf(stderr,"mod_prov: HANDLER call mod_prov_method_handler\n");
	fflush(stderr);

    const char *script;
//    const char *command;

	// We need to flush the stream so that the message appears right away.
	// Performing an fflush() in a production system is not good for
	// performance - don't do this for real.

	// Start with the real stuff
	strProv_config_rec *conf = (strProv_config_rec *) ap_get_module_config(r->per_dir_config, &prov_module);
	strProv_module_config_rec *mod_conf = (strProv_module_config_rec *) ap_get_module_config(r->server->module_config, &prov_module);

    // Uit te voeren script
    script=conf->szProv_script;
    if (script == NULL) {
    	fprintf(stderr,"mod_prov: ERROR: Script is empty\n");
    	fflush(stderr);
    	return DECLINED;
    }

    // Check if we need to check the cookie
	fprintf(stderr,"mod_prov: Value of checkcookie %i\n",conf->nProv_chkcookie);
	fflush(stderr);
    if (conf->nProv_chkcookie) {
    	// We need to check the cookie
    	fprintf(stderr,"mod_prov: Need to check cookie\n");
    	fflush(stderr);

		// Check if cookie is set. Execute command only if correct cookie is NOT set
		char *szCookieValue=NULL;
		unless(szCookieValue = extract_cookie(r, conf->szProv_cookiename))
		{
			// Even though we need to check the cookie, it was not found
			// Cannot check for valid session, let request through
			fprintf(stderr,"mod_prov: Cookie not found. Was looking for cookiename %s\n",conf->szProv_cookiename);
			fflush(stderr);
			// We should make this an error, as this should not happen (see design)
			return DECLINED;	//Always return DECLINED
		}
		fprintf(stderr,"mod_prov: Cookie found. Value is %s\n",szCookieValue);
		fprintf(stderr,"mod_prov: CACHE: Checking cookie value in cache\n");
		fprintf(stderr,"mod_prov: CACHE: For now, just testing\n");
		fflush(stderr);

		/******************************************
		 * BEGIN: Testing and printing the cache
		 ******************************************/
		// Fill a cache element with current time and dummy cookie value
		// Note that the creation of this cache element should be done much later, after
		// the run_program is run successfully
/*
		fprintf(stderr,"mod_prov: CACHE: defined element\n");
		fprintf(stderr,"mod_prov: size of Prov_cache_element is %d bytes\n",(int)sizeof(Prov_cache_element));
		fflush(stderr);


		Prov_cache_element *el;
		el = elt_new(conf);
		//el = (Prov_cache_element *) apr_pcalloc(conf->Prov_cache_pool, sizeof(Prov_cache_element));

		fprintf(stderr,"mod_prov: CACHE: allocated memory for cache element\n");
		fflush(stderr);
		//APR_RING_ELEM_INIT(el,link); //XXX: misschien lost dit het loopen op?
		el->entry_time=apr_time_now();
		el->cookie_value="test";
		fprintf(stderr,"mod_prov: CACHE: filled cache element with test values\n");
		fflush(stderr);
		APR_RING_INSERT_HEAD(conf->Prov_cache_list_head, el, Prov_cache_element, link);
		fprintf(stderr,"mod_prov: CACHE: inserted element in cache\n");
		fflush(stderr);

		//Iterate over cache to print content (for clarity, code implemented here
		// does the same as APR_RING_FOREACH)
		Prov_cache_element *ep = NULL;
//		ep = (Prov_cache_element *) apr_pcalloc(conf->Prov_cache_pool, sizeof(Prov_cache_element));


/*		ep = APR_RING_FIRST(conf->Prov_cache_list_head);

		fprintf(stderr,"mod_prov: CACHE: Entry 1 in cache is:\n");
		fprintf(stderr,"mod_prov:           cookie_value: %s\n", ep->cookie_value);
		fprintf(stderr,"mod_prov:           entry_time: %" APR_TIME_T_FMT "[us]\n", ep->entry_time);
		fflush(stderr);
	 	ep = APR_RING_NEXT(ep, link);
	 	//ep = APR_RING_LAST(conf->Prov_cache_list_head);
		fprintf(stderr,"mod_prov: CACHE: Entry 2 in cache is:\n");
		fprintf(stderr,"mod_prov:           cookie_value: %s\n", ep->cookie_value);
		fprintf(stderr,"mod_prov:           entry_time: %" APR_TIME_T_FMT "[us]\n", ep->entry_time);
		fflush(stderr);
*/
// FIXME: This gives an infinite loop..

		Prov_cache_element *ep = NULL;
		for (ep = APR_RING_FIRST(mod_conf->cache_list_head);
				   ep != APR_RING_SENTINEL(mod_conf->cache_list_head, Prov_cache_element, link);
				   ep = APR_RING_NEXT(ep, link)) {
			fprintf(stderr,"mod_prov: CACHE: Entry in cache is:\n");
			fprintf(stderr,"mod_prov:           cookie_value: %s\n", ep->cookie_value);
			fprintf(stderr,"mod_prov:           entry_time: %" APR_TIME_T_FMT "[us]\n", ep->entry_time);
			fflush(stderr);
			}


/*
		while (ep != sentinel) {
				fprintf(stderr,"mod_prov: CACHE: Entry in cache is:\n");
				fprintf(stderr,"mod_prov:           cookie_value: %s\n", ep->cookie_value);
				fprintf(stderr,"mod_prov:           entry_time: %" APR_TIME_T_FMT "[us]\n", ep->entry_time);
				fflush(stderr);
				ep = APR_RING_NEXT(ep, link);

		}
*/

		/******************************************
		 * END: Testing and printing the cache
		 ******************************************/

	 	/**********************************************
	 	 * Now we know the cookie with the correct name is set, now
	 	 * we look for the cookievalue in the cache
	 	 **********************************************/
/*	Disable this untill all caching things have been worken out correctly
 */
		fprintf(stderr,"mod_prov: CACHE: Looking for cookie value\n");
		fflush(stderr);

		Prov_cache_element *el_tmp;
		el_tmp = (Prov_cache_element *) apr_pcalloc(mod_conf->cache_pool, sizeof(Prov_cache_element));

		el_tmp = APR_RING_FIRST(mod_conf->cache_list_head);

		while (el_tmp != APR_RING_SENTINEL(mod_conf->cache_list_head, Prov_cache_element, link)) {
			if (apr_strnatcmp(el_tmp->cookie_value, szCookieValue) == 0) {
				// Found cookie in cache
				fprintf(stderr,"mod_prov: CACHE: Found cookie value %s in cache\n", szCookieValue);
				fflush(stderr);
				return DECLINED;
			}
			// Cookie value not in cache
			fprintf(stderr,"mod_prov: CACHE: Element does not contain cookie value %s\n", szCookieValue);
			fflush(stderr);
			el_tmp = APR_RING_NEXT(el_tmp, link);
		}

		// If we get to here, cookie value is not found in cache
		fprintf(stderr,"mod_prov: CACHE: Cookie value %s not found in cache\n", szCookieValue);
		fflush(stderr);

		Prov_cache_element *el;
		el = elt_new(mod_conf);
		//el = (Prov_cache_element *) apr_pcalloc(conf->Prov_cache_pool, sizeof(Prov_cache_element));

		fprintf(stderr,"mod_prov: CACHE: allocated memory for cache element\n");
		fflush(stderr);
		el->entry_time=apr_time_now();
		el->cookie_value=apr_pstrdup(mod_conf->cache_pool, szCookieValue);
		fprintf(stderr,"mod_prov: CACHE: filled cache element with test values\n");
		fflush(stderr);
		APR_RING_INSERT_HEAD(mod_conf->cache_list_head, el, Prov_cache_element, link);
		fprintf(stderr,"mod_prov: CACHE: inserted element in cache\n");
		fflush(stderr);

    } else {
    	// No need to check the cookie
		fprintf(stderr,"mod_prov: Not necessary to check cookie\n");
		fflush(stderr);
    }


    /*****************************************************
     *
     * Prepare call to script
     *
     *****************************************************/

    fprintf(stderr,"mod_prov: Preparing to call script\n");
    fflush(stderr);

    apr_status_t rv;
    apr_procattr_t *pattr;

    //apr_initialize();				// Hoeft waarschijnlijk niet voor een module
    //apr_pool_create(&mp, NULL);	// Wij hoeven geen pool te maken

    /* prepare process attribute */
    if ((rv = apr_procattr_create(&pattr, r->pool)) != APR_SUCCESS) {
        fprintf(stderr,"mod_prov: Error preparing script - apr_procattr_create\n");
        fprintf(stderr,"mod_prov: Error: %d\n",rv);
        fflush(stderr);
    }
    if ((rv = apr_procattr_io_set(pattr, APR_NO_PIPE, APR_NO_PIPE, APR_NO_PIPE)) != APR_SUCCESS) {
        fprintf(stderr,"mod_prov: Error preparing script - apr_procattr_io_set\n");
        fprintf(stderr,"mod_prov: Error: %d\n",rv);
        fflush(stderr);
    }
    /* default value: APR_PROGRAM */

//    if ((rv = apr_procattr_cmdtype_set(pattr, APR_PROGRAM_ENV)) != APR_SUCCESS) {
/*
    if ((rv = apr_procattr_cmdtype_set(pattr, APR_PROGRAM)) != APR_SUCCESS) {
        fprintf(stderr,"mod_prov: Error preparing script - apr_procattr_cmdtype_set\n");
        fprintf(stderr,"mod_prov: Error: %d\n",rv);
        fflush(stderr);
    }
*/
    /* detaching process has different effects on platform.
     * On Unix, detached process implies a daemon process.
     * On Windows, detached process implies a process without the console window. */
/*
    apr_procattr_detach_set(pattr, TRUE);

    apr_table_do(print_iterate_func, r, r->subprocess_env, NULL);
*/

    char **env = ap_create_environment(r->pool, r->subprocess_env);

    /* run the program */
    run_program(script, pattr, r->pool, (const char* const *)env);

    fprintf(stderr,"mod_prov: End of call to script\n");
    fflush(stderr);

    if (conf->nProv_chkcookie) {
    	// Because we need to handle cookies, we need to put the
    	// cookie value in the cache and end
    	// FIXME: this should be implemented
    	return DECLINED;
    }

	// Return DECLINED so that the Apache core will keep looking for
	// other modules to handle this request.  This effectively makes
	// this module completely transparent.
    return DECLINED;
}

/*
 * This function is a callback and it declares what other functions
 * should be called for request processing and configuration requests.
 * This callback function declares the Handlers for other events.
 */
static void mod_prov_register_hooks (apr_pool_t *p)
{
	fprintf(stderr,"mod_prov: HANDLER call mod_prov_register_hooks\n");
	fflush(stderr);

	// I think this is the call to make to register a handler for method calls (GET PUT et. al.).
	// We will ask to be last so that the comment has a higher tendency to
	// go at the end.

	// FIXME: Waarschijnlijk moet dit ap_hook_auth_checker worden, maar deze werkt alleen
	// bij een require directive, dus voor testen niet zo geschikt
	//ap_hook_handler(mod_prov_method_handler, NULL, NULL, APR_HOOK_LAST);
//	ap_hook_handler(mod_prov_method_handler, NULL, NULL, APR_HOOK_FIRST);
	ap_hook_fixups(mod_prov_method_handler, NULL, NULL, APR_HOOK_LAST);
}

/***************************************************************************************
 * Functions to set configuration directives
 *
 * For definitions of parameters passed to these functions,
 * see http_config.h in "typedef union"
 ***************************************************************************************/
static const char *add_provscript(cmd_parms *parms,
  void *mconfig, const char *arg)
{
	fprintf(stderr,"mod_prov: FUNCTION directive call add_provscript\n");
	fflush(stderr);
	strProv_config_rec *conf = (strProv_config_rec *)mconfig;
	conf->szProv_script = (char *) arg;
	return NULL;
}

static const char *add_provcookiename(cmd_parms *parms,
  void *mconfig, const char *arg)
{
	fprintf(stderr,"mod_prov: FUNCTION directive call add_provcookiename\n");
	fflush(stderr);
	strProv_config_rec *conf = (strProv_config_rec *)mconfig;
	conf->szProv_cookiename = (char *) arg;
	return NULL;
}


static const char *add_provchkcookie(cmd_parms *parms,
  void *mconfig, int on)
{
	fprintf(stderr,"mod_prov: FUNCTION directive call add_provchkcookie\n");
	fflush(stderr);
	strProv_config_rec *conf = (strProv_config_rec *)mconfig;
	conf->nProv_chkcookie = on;
	return NULL;
}

/***************************************************************************************
 * A declaration of the configuration directives that are supported by this module.
 ***************************************************************************************/

static void *create_mod_prov_dir_config(apr_pool_t *p, char *d)
{
	fprintf(stderr,"mod_prov: HANDLER call create_mod_prov_dir_config\n");
	fflush(stderr);

    strProv_config_rec *conf = apr_palloc(p, sizeof(*conf));

    // Set default values for the module directives
    conf->szProv_script = apr_pstrdup(p,"/home/hansz/coin/trunk/Adobe Connect/prov_script/prov_script.sh");
    conf->szProv_cookiename = apr_pstrdup(p,"_shibsession_54cf5cf957fe0f3895da181e8a390ae398a19bdd");
    conf->nProv_chkcookie = 1;  /* default is to check the cookie */

    return conf;
}

/***************************************************************
 *
 * Creates the per-server configuration records.
 *
 ***************************************************************/
static void *create_mod_prov_config(apr_pool_t *p, server_rec *s)
{
	fprintf(stderr,"mod_prov: HANDLER call create_mod_prov_config\n");
	fflush(stderr);

	strProv_module_config_rec *conf = apr_palloc(p, sizeof(*conf));
    
	apr_pool_create(&conf->cache_pool, p);
	conf->cache_list_head = apr_palloc(conf->cache_pool, sizeof(struct Prov_cache_list));
	APR_RING_INIT(conf->cache_list_head, Prov_cache_element, link);

	return (void *) conf;
}

/* apache config directives for the module */
static const command_rec mod_prov_cmds[] =
{
	//TAKE1 betekent dat maar 1 parameter aan directive wordt gegeven
	// Zie ook http_config.h voor definitie van deze macros

	AP_INIT_TAKE1("ModuleProvScript", add_provscript, NULL,
     OR_AUTHCFG, "Full path to location of script to call"),
    AP_INIT_TAKE1("ModuleProvCookieName", add_provcookiename, NULL,
     OR_AUTHCFG, "Cookie name to check before script is run"),
    AP_INIT_FLAG ("ModuleProvChkCookie", add_provchkcookie, NULL,
     OR_AUTHCFG, "Set to 'Off' to not check for a cookie before the script is run... On by default"),
    {NULL}

};

/*
 * Declare and populate the module's data structure.  The
 * name of this structure ('prov_module') is important - it
 * must match the name of the module.  This structure is the
 * only "glue" between the httpd core and the module.
 */
module AP_MODULE_DECLARE_DATA prov_module =
{
	// Only one callback function is provided.  Real
	// modules will need to declare callback functions for
	// server/directory configuration, configuration merging
	// and other tasks.
	STANDARD20_MODULE_STUFF,
	create_mod_prov_dir_config,			/* per-directory config creator */
	NULL,								/* dir config merger */
	create_mod_prov_config,				/* server config creator, called once to create initial module configuration */
	NULL,								/* server config merger */
	mod_prov_cmds,						/* configuration directive handlers */
	mod_prov_register_hooks,			/* callback for registering hooks */
};
